/*
	Program ..... XW
	File Info ... Read README and LICENSE
*/

/*-
 * Copyright (c) 1980, 1991, 1993, 1994
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* This source was mostly taken from w.c, I made a lots of changes and 
   added source from a few other freebsd files that w.c requries to compile
*/

#include "global.h"

#include <sys/param.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/tty.h>
#include <sys/types.h>
#include <sys/resource.h>

#include <machine/cpu.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <kvm.h>
#include <netdb.h>
#include <nlist.h>
#include <paths.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <utmp.h>
#include <vis.h>
#include <locale.h>

#include <arpa/nameser.h>
#include <resolv.h>

enum type { CHAR, UCHAR, SHORT, USHORT, INT, UINT, LONG, ULONG, KPTR };

struct usave {
        struct  timeval u_start;
        struct  rusage u_ru;
        struct  rusage u_cru;
        char    u_acflag;
        char    u_valid;
};

#define KI_PROC(ki) (&(ki)->ki_p->kp_proc)
#define KI_EPROC(ki) (&(ki)->ki_p->kp_eproc)

typedef struct kinfo {
        struct kinfo_proc *ki_p;	/* proc structure */
        struct usave ki_u;		/* interesting parts of user */
        char *ki_args;			/* exec args */
        char *ki_env;			/* environment */
} KINFO;

struct timeval	boottime;
struct utmp	utmp;
struct winsize	ws;
kvm_t	       *kd;
time_t		now;			/* the current time of day */
time_t		uptime;			/* time of last reboot & elapsed time since */
int		argwidth;		/* width of tty */
int		header = 1;		/* true if -h flag: don't print heading */
int		nflag;			/* true if -n flag: don't convert addrs */
int		dflag;			/* true if -d flag: output debug info */
int		sortidle;		/* sort by idle time */
char	      **sel_users;		/* login array of particular users selected */
char		domain[MAXHOSTNAMELEN];

static char *cmdpart __P((char *));
static char *shquote __P((char **));

#define debugproc(p) *((struct kinfo_proc **)&(p)->kp_eproc.e_spare[0])

#define ISRUN(p)        (((p)->p_stat == SRUN) || ((p)->p_stat == SIDL))
#define TESTAB(a, b)    ((a)<<1 | (b))
#define ONLYA   2
#define ONLYB   1
#define BOTH    3

static struct stat	*ttystat __P((char *, int));
struct proc;
void    pr_attime __P((time_t *, time_t *));
int     pr_idle __P((time_t));
int     proc_compare __P((struct proc *, struct proc *));

char *fmt_argv __P((char **, char *, int));     /* ../../bin/ps/fmt.c */

static char *
shquote(argv)
        char **argv;
{
        long arg_max;
        char **p, *dst, *src;
        static char *buf = NULL;

        if (buf == NULL) {
                if ((arg_max = sysconf(_SC_ARG_MAX)) == -1)
                        errx(1, "sysconf _SC_ARG_MAX failed");
                if ((buf = malloc((4 * arg_max)  +  1)) == NULL)
                        errx(1, "malloc failed");
        }

        if (*argv == 0) {
                buf[0] = 0;
                return (buf);
        }
        dst = buf;
        for (p = argv; (src = *p++) != 0; ) {
                if (*src == 0)
                        continue;
                strvis(dst, src, VIS_NL | VIS_CSTYLE);
                while (*dst)
                        dst++;
                *dst++ = ' ';
        }
        /* Chop off trailing space */
        if (dst != buf)
                dst--;
        *dst = '\0';
        return (buf);
}

static char *
cmdpart(arg0)
        char *arg0;
{
        char *cp;

        return ((cp = strrchr(arg0, '/')) != NULL ? cp + 1 : arg0);
}

char *
fmt_argv(argv, cmd, maxlen)
        char **argv;
        char *cmd;
        int maxlen;
{
        int len;
        char *ap, *cp;

        if (argv == 0 || argv[0] == 0) {
                if (cmd == NULL)
                        return ("");
                ap = NULL;
                len = maxlen + 3;
        } else {
                ap = shquote(argv);
                len = strlen(ap) + maxlen + 4;
        }
        if ((cp = malloc(len)) == NULL)
                return (NULL);
        if (ap == NULL)
                sprintf(cp, " (%.*s)", maxlen, cmd);
        else if (strncmp(cmdpart(argv[0]), cmd, maxlen) != 0)
                sprintf(cp, "%s (%.*s)", ap, maxlen, cmd);
        else
                (void) strcpy(cp, ap);
        return (cp);
}

void
pr_attime(started, now)
        time_t *started, *now;
{
        static char buf[256];
        struct tm tp, tm;
        time_t diff;
        char fmt[20];

        tp = *localtime(started);
        tm = *localtime(now);
        diff = *now - *started;

        /* If more than a week, use day-month-year. */
        if (diff > 86400 * 7)
                (void)strcpy(fmt, "%d%b%y");

        /* If not today, use day-hour-am/pm. */
        else if (tm.tm_mday != tp.tm_mday ||
                 tm.tm_mon != tp.tm_mon ||
                 tm.tm_year != tp.tm_year) {
        /* The line below does not take DST into consideration */
        /* else if (*now / 86400 != *started / 86400) { */
                (void)strcpy(fmt, __CONCAT("%a%", "I%p"));
        }

        /* Default is hh:mm{am,pm}. */
        else {
                (void)strcpy(fmt, __CONCAT("%l:%", "M%p"));
        }

        (void)strftime(buf, sizeof(buf) - 1, fmt, &tp);
        buf[sizeof(buf) - 1] = '\0';
        (void)sprintf(userlist->signon, "%s", buf);
}

int
pr_idle(idle)
        time_t idle;
{
	if (idle >= 36 * 3600)
	{
		int days = idle / 86400;
		(void)sprintf(userlist->idle, " %dday%s ", days, days > 1 ? "s" : " " );

		if (days >= 100)
			return (2);
		if (days >= 10)
			return (1);
	}
	else if (idle >= 3600)
		(void)sprintf(userlist->idle, " %2d:%02d ", (int)(idle / 3600), (int)((idle % 3600) / 60));
	else if (idle / 60 == 0)
		(void)sprintf(userlist->idle, "     - ");
	else
		(void)sprintf(userlist->idle, "    %2d ", (int)(idle / 60));

	return (0);
}


int
proc_compare(p1, p2)
        register struct proc *p1, *p2;
{

	if (p1 == NULL)
		return (1);

	switch (TESTAB(ISRUN(p1), ISRUN(p2))) {
		case ONLYA:
			return (0);
		case ONLYB:
			return (1);
		case BOTH:
			if (p2->p_estcpu > p1->p_estcpu)
				return (1);
			if (p1->p_estcpu > p2->p_estcpu)
				return (0);
			return (p2->p_pid > p1->p_pid);
	}

	switch (TESTAB(p1->p_stat == SZOMB, p2->p_stat == SZOMB)) {
		case ONLYA:
			return (1);
		case ONLYB:
			return (0);
		case BOTH:
			return (p2->p_pid > p1->p_pid); /* tie - return highest pid */
	}

	if (p2->p_slptime > p1->p_slptime)
		return (0);
	if (p1->p_slptime > p2->p_slptime)
		return (1);
	if (p1->p_flag & P_SINTR && (p2->p_flag & P_SINTR) == 0)
		return (1);
	if (p2->p_flag & P_SINTR && (p1->p_flag & P_SINTR) == 0)
		return (0);

	return (p2->p_pid > p1->p_pid);
}

static struct stat *
ttystat(line, sz)
	char *line;
	int sz;
{
	static struct stat sb;
	char ttybuf[MAXPATHLEN];

	(void)snprintf(ttybuf, sizeof(ttybuf), "%s%.*s", _PATH_DEV, sz, line);
	if (stat(ttybuf, &sb))
	{
		warn("%s", ttybuf);
		return (NULL);
	}

	return (&sb);
}

void do_it()
{
        struct kinfo_proc *kp;
        struct kinfo_proc *dkp;
        struct hostent *hp;
        struct stat *stp;
        FILE *ut;
        u_long l;
        time_t touched;
        int i, nentries, nusers, wcmd=1, longidle;
        char *memf, *nlistf, *p, *x;
        char buf[MAXHOSTNAMELEN], errbuf[256];
	int append_it, previous;

	struct  entry {
        	struct  entry *next;
        	struct  utmp utmp;
        	dev_t   tdev;                   /* dev_t of terminal */
        	time_t  idle;                   /* idle time of terminal in seconds */
        	struct  kinfo_proc *kp;         /* `most interesting' proc */
        	char    *args;                  /* arg list of interesting process */
        	struct  kinfo_proc *dkp;        /* debug option proc list */
	} *ep, *ehead = NULL, **nextp = &ehead;

        (void)setlocale(LC_ALL, "");

        memf = nlistf = _PATH_DEVNULL;

        if (!(_res.options & RES_INIT))
                res_init();
        _res.retrans = 2;       /* resolver timeout to 2 seconds per try */
        _res.retry = 1;         /* only try once.. */

        if ((kd = kvm_openfiles(nlistf, memf, NULL, O_RDONLY, errbuf)) == NULL)
                errx(1, "%s", errbuf);

        (void)time(&now);
        if ((ut = fopen(_PATH_UTMP, "r")) == NULL)
                err(1, "%s", _PATH_UTMP);

	for (nusers = 0; fread(&utmp, sizeof(utmp), 1, ut);) {
		if (utmp.ut_name[0] == '\0')
			continue;
		if (!(stp = ttystat(utmp.ut_line, UT_LINESIZE)))
			continue;	/* corrupted record */
		++nusers;
		if (wcmd == 0)
			continue;
		if (sel_users) {
			int usermatch;
			char **user;

			usermatch = 0;
			for (user = sel_users; !usermatch && *user; user++)
				if (!strncmp(utmp.ut_name, *user, UT_NAMESIZE))
					usermatch = 1;
			if (!usermatch)
				continue;
		}
		if ((ep = calloc(1, sizeof(struct entry))) == NULL)
			errx(1, "calloc");
		*nextp = ep;
		nextp = &ep->next;
		memmove(&ep->utmp, &utmp, sizeof(struct utmp));
		ep->tdev = stp->st_rdev;
#ifdef CPU_CONSDEV
		/*
		 * If this is the console device, attempt to ascertain
		 * the true console device dev_t.
		 */
		if (ep->tdev == 0) {
			int mib[2];
			size_t size;

			mib[0] = CTL_MACHDEP;
			mib[1] = CPU_CONSDEV;
			size = sizeof(dev_t);
			(void)sysctl(mib, 2, &ep->tdev, &size, NULL, 0);
		}
#endif
		touched = stp->st_atime;
		if (touched < ep->utmp.ut_time) {
			/* tty untouched since before login */
			touched = ep->utmp.ut_time;
		}
		if ((ep->idle = now - touched) < 0)
			ep->idle = 0;
	}
	(void)fclose(ut);

	if ((kp = kvm_getprocs(kd, KERN_PROC_ALL, 0, &nentries)) == NULL)
		err(1, "%s", kvm_geterr(kd));
	for (i = 0; i < nentries; i++, kp++) {
		struct proc *pr = &kp->kp_proc;
		struct eproc *e;

		if (pr->p_stat == SIDL || pr->p_stat == SZOMB)
			continue;
		e = &kp->kp_eproc;
		for (ep = ehead; ep != NULL; ep = ep->next) {
			if (ep->tdev == e->e_tdev) {
				/*
				 * proc is associated with this terminal
				 */
				if (ep->kp == NULL && e->e_pgid == e->e_tpgid) {
					/*
					 * Proc is 'most interesting'
					 */
					if (proc_compare(&ep->kp->kp_proc, pr))
						ep->kp = kp;
				}
				/*
				 * Proc debug option info; add to debug
				 * list using kinfo_proc kp_eproc.e_spare
				 * as next pointer; ptr to ptr avoids the
				 * ptr = long assumption.
				 */
				dkp = ep->dkp;
				ep->dkp = kp;
				debugproc(kp) = dkp;
			}
		}
	}
	argwidth = 80;
	if (argwidth < 4)
		argwidth = 8;
	for (ep = ehead; ep != NULL; ep = ep->next) {
		if (ep->kp == NULL) {
			ep->args = "-";
			continue;
		}
		ep->args = fmt_argv(kvm_getargv(kd, ep->kp, argwidth),
		    ep->kp->kp_proc.p_comm, MAXCOMLEN);

		if (ep->args == NULL)
			err(1, NULL);
	}
	/* sort by idle time */
	if (sortidle && ehead != NULL) {
		struct entry *from, *save;

		from = ehead;
		ehead = NULL;
		while (from != NULL) {
			for (nextp = &ehead;
			    (*nextp) && from->idle >= (*nextp)->idle;
			    nextp = &(*nextp)->next)
				continue;
			save = from;
			from = from->next;
			save->next = *nextp;
			*nextp = save;
		}
	}

	if (!nflag) {
		if (gethostname(domain, sizeof(domain) - 1) < 0 ||
		    (p = strchr(domain, '.')) == NULL)
			domain[0] = '\0';
		else {
			domain[sizeof(domain) - 1] = '\0';
			memmove(domain, p, strlen(p) + 1);
		}
	}

	for (ep = ehead; ep != NULL; ep = ep->next) {
		char host_buf[UT_HOSTSIZE + 1];

		host_buf[UT_HOSTSIZE] = '\0';
		strncpy(host_buf, ep->utmp.ut_host, UT_HOSTSIZE);
		p = *host_buf ? host_buf : "-";
		if ((x = strchr(p, ':')) != NULL)
			*x++ = '\0';
		if (!nflag && isdigit(*p) && (l = inet_addr(p)) != -1 &&
		    (hp = gethostbyaddr((char *)&l, sizeof(l), AF_INET))) {
			if (domain[0] != '\0') {
				p = hp->h_name;
				p += strlen(hp->h_name);
				p -= strlen(domain);
				if (p > hp->h_name &&
				    strcasecmp(p, domain) == 0)
					*p = '\0';
			}
			p = hp->h_name;
		}
		if (nflag && *p && strcmp(p, "-") &&
		    inet_addr(p) == INADDR_NONE) {
			hp = gethostbyname(p);

			if (hp != NULL) {
				struct in_addr in;

				memmove(&in, hp->h_addr, sizeof(in));
				p = inet_ntoa(in);
			}
		}
		if (x) {
			(void)snprintf(buf, sizeof(buf), "%s:%s", p, x);
			p = buf;
		}
		if (dflag) {
			for (dkp = ep->dkp; dkp != NULL; dkp = debugproc(dkp)) {
				char *ptr;

				ptr = fmt_argv(kvm_getargv(kd, dkp, argwidth),
				    dkp->kp_proc.p_comm, MAXCOMLEN);

				if (ptr == NULL)
					ptr = "-";
				(void)printf("\t\t%-9d %s\n",
				    dkp->kp_proc.p_pid, ptr);
			}
		}

/*		(void)printf("%-*.*s %-*.*s %-*.*s ",
		    UT_NAMESIZE, UT_NAMESIZE, ep->utmp.ut_name,
		    UT_LINESIZE, UT_LINESIZE,
		    strncmp(ep->utmp.ut_line, "tty", 3) &&
		    strncmp(ep->utmp.ut_line, "cua", 3) ?
		    ep->utmp.ut_line : ep->utmp.ut_line + 3,
		    UT_HOSTSIZE, UT_HOSTSIZE, *p ? p : "-");
*/
		pr_attime(&ep->utmp.ut_time, &now);
		longidle = pr_idle(ep->idle);
//		(void)printf("%.*s\n", argwidth - longidle, ep->args);

		sprintf(userlist->username, "%s", ep->utmp.ut_name);
		sprintf(userlist->tty, "%s", ep->utmp.ut_line);
		sprintf(userlist->hostname, "%s", *p ? p : "-");

		append_it = 1;
		for (previous = 0; previous < row_total; previous++)
		{
			list_info(1, previous);
			if (strcmp(userlist->tty, return_info) == 0)
			{
				list_info(0, previous);
				if (strcmp(userlist->username, return_info) == 0 )
					append_it = 0;
			}
		}

		if ( append_it == 1 )
			xw_clist_append(userlist);

		userlist->next = (struct __userlist *) malloc((unsigned) sizeof(struct __userlist));
		userlist = userlist->next;
	}

	userlist->next = NULL;
	kvm_close(kd);
	ep = NULL;

	return;
}



